package lichen.utils
import java.util.concurrent.TimeUnit
import java.util.Date

import org.slf4j.LoggerFactory
import scala.actors.{Actor, Exit, TIMEOUT}
import org.apache.tapestry5.ioc.internal.services.cron.CronExpression

/**
 * lightweight scheduler using scala reactWithin timeout pattern
 */
object Scheduler {

    /**
     * Stopper trait
     */
    trait Stopper {
        def stop
    }

    private val logger = LoggerFactory.getLogger(getClass)

    //scheuler start mode
    abstract class SchedulerStartMode

    //start at once
    case object StartAtOnce extends SchedulerStartMode

    //start at delay
    case object StartAtDelay extends SchedulerStartMode

    /**
     * 延迟执行某项任务
     * @param f 待执行的函数
     * @param time 延迟时间长度
     * @param unit 延迟的时间单位
     */
    def delayExecuteTask(f: => Unit, time: Long, unit: TimeUnit): Unit = {
        import Actor._
        val timeOut = unit.toMillis(time);
        actor {
            reactWithin(timeOut) {
                case TIMEOUT =>
                    //execute function
                    try {
                        f;
                    } catch {
                        case e: Throwable => logger.warn(e.toString, e)
                    };
                case _ =>
                    logger.warn("receive other some messsage")
            }
        }
    }

    /**
     * 定时执行某一项任务
     * @param f 定时执行的函数
     * @param exp 定时器的配置
     * @param startMode 启动的模式 StartAtOnce StartAtDelay
     * @return 定时的对象
     * @see CronExpression
     */
    def schedule(f: => Unit, exp: String, startMode: SchedulerStartMode) = new AnyRef with Stopper {

        import Actor._

        private val expression = new CronExpression(exp)
        private val a = actor {
            val waitingTime: Long = startMode match {
                //delay to start work
                case StartAtDelay =>
                    calculateNextTime(expression)
                case _ =>
                    //default one minute
                    10000
            }
            loop(waitingTime)
        }

        private def loop(time: Long): Unit = reactWithin(time) {
            case TIMEOUT =>
                //execute function
                try {
                    f
                } catch {
                    case e: Throwable => logger.warn(e.toString, e)
                };
                //loop next time
                loop(calculateNextTime(expression))
            case Exit =>
                reply();
                exit()
        }

        /**
         * stop the scheduler
         */
        override def stop {
            logger.info("stop scheduler")
            a !? Exit
        }
    }

    private[utils] def calculateNextTime(expression: CronExpression): Long = {
        val now = new Date()
        val nextTime = expression.getNextValidTimeAfter(now)
        logger.debug("nextTime:{}", nextTime)
        nextTime.getTime - now.getTime
    }
}