package lichen.utils

import javax.xml.bind.{Marshaller, JAXBContext}
import javax.xml.bind.util.ValidationEventCollector
import javax.xml.validation.SchemaFactory
import javax.xml.XMLConstants
import javax.xml.transform.stream.StreamSource
import java.io._
import io.{Codec, Source}

/**
 * 使用Jaxb方式解析XML文件
 *
 * @author jcai
 * @version 0.1
 */
object XmlLoader {
    /**
     * 加载某一xml字符串
     * @param content XML内容
     * @param symbols 一些需要需要解析的名值对
     * @param encoding 编码
     * @param xsd xsd校验文件
     * @param m manifest
     * @tparam T 返回的类
     * @return 解析后的类
     */
    def loadXmlString[T <: Object](content: String,
                              symbols: Option[Map[String, String]] = None, 
                              encoding: Codec = Codec(Codec.UTF8),
                              xsd: Option[InputStream] = None)(implicit m: Manifest[T]): T = {
        val xml =  SymbolExpander.expand(content, symbols)
        loadXml(new ByteArrayInputStream(xml.getBytes(encoding.name)),encoding,xsd)
    }

    /**
     * 解析xml文件
     * @param is 输入流
     * @param encoding 编码
     * @param xsd xsd文件
     * @param m manifest定义
     * @tparam T 返回的对象
     * @return 解析后的对象
     */
    def loadXml[T <: Object](is: InputStream,
                             encoding: Codec=Codec(Codec.UTF8),
                             xsd: Option[InputStream]=None)(implicit m: Manifest[T]): T = {
        val vec = new ValidationEventCollector();
        try {
            //obtain type parameter
            val clazz = m.erasure.asInstanceOf[Class[T]]
            //create io reader
            val reader = new InputStreamReader(is, encoding.name);
            val context = JAXBContext.newInstance(clazz)
            //unmarshal xml
            val unmarshaller = context.createUnmarshaller()
            //.unmarshal(reader).asInstanceOf[T]
            if (xsd.isDefined) {
                val sf = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);
                val schemaSource = new StreamSource(xsd.get,encoding.name);
                val schema = sf.newSchema(schemaSource)
                unmarshaller.setSchema(schema)
                unmarshaller.setEventHandler(vec)
            }
            unmarshaller.unmarshal(reader).asInstanceOf[T]
        } finally {
            close(is)
            if (xsd.isDefined)
                close(xsd.get)
            if (vec.hasEvents) {
                val veOption = vec.getEvents.headOption
                if (veOption.isDefined) {
                    val ve = veOption.get
                    val vel = ve.getLocator
                    throw new IllegalArgumentException("line %s column %s :%s".
                      format(vel.getLineNumber, vel.getColumnNumber, ve.getMessage))
                }
            }
        }
    }

    private def close(io: Closeable) {
        try {
            io.close
        } catch {
            case e =>
        }
    }

    /**
     * 把对象转换为xml文档
     * @param obj 待转换的对象
     * @param encoding 文档编码
     * @return 转换后的xml内容
     */
    def toXml(obj: Any, encoding: Codec = Codec(Codec.UTF8)): String = {
        val context = JAXBContext.newInstance(obj.getClass)
        val marshaller = context.createMarshaller()
        marshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, true);
        marshaller.setProperty(Marshaller.JAXB_ENCODING, encoding.name);
        val out = new ByteArrayOutputStream
        marshaller.marshal(obj, out)
        new String(out.toByteArray, encoding.name)
    }
}