package lichen.utils

// Copyright 2012 The EGF IT Software Department.
// site: http://www.egfit.com

import org.apache.zookeeper.ZooDefs.Ids
import collection.JavaConversions._
import org.apache.zookeeper.{WatchedEvent, Watcher, CreateMode, ZooKeeper}
import org.apache.zookeeper.data.Stat
import org.apache.zookeeper.KeeperException.{NoNodeException, ConnectionLossException, NodeExistsException}
import org.slf4j.LoggerFactory
import org.apache.zookeeper.Watcher.Event.{KeeperState, EventType}
import java.util.concurrent.{ConcurrentHashMap, CountDownLatch}
import org.apache.zookeeper._
import java.io.Closeable
import java.util.concurrent.atomic.AtomicInteger
import org.apache.zookeeper.AsyncCallback.{Children2Callback, DataCallback, ChildrenCallback}
import server.DataTree


/**
 * zookeeper的模板类,提供对zookeeper的一些操作
 * @author jcai
 */
class ZookeeperTemplate(address: String,
                        basePath: Option[String] = None,
                        sessionTimeout: Int = 5000,
                        connectedFun: (ZookeeperTemplate) => Unit = {
                            zk =>
                        }) extends Closeable {
    //禁止zookeeper自动reset watcher，在本类里面统一处理
    //ClientCnxn.setDisableAutoResetWatch(true)

    private val logger = LoggerFactory getLogger getClass
    /**add base path **/
    private val serverAddress = address + basePath.getOrElse("")
    private var zookeeper: ZooKeeper = null
    private var connectionLatch: CountDownLatch = null
    private var sessionExpiredFlag: AtomicInteger = new AtomicInteger(0)
    private val watchers = new ConcurrentHashMap[String, (ZooKeeper) => Any]()
    //do connect
    connect

    /**
     * connect() attaches to the remote zookeeper and sets an instance variable.
     */
    private def connect {
        if (connectionLatch == null || connectionLatch.getCount != 1) {
            connectionLatch = new CountDownLatch(1)
        }
        val assignLatch = new CountDownLatch(1)

        if (zookeeper != null) {
            try {
                zookeeper.close()
            } catch {
                case e =>
                    logger.warn("Fail to close client", e)
            }
            zookeeper = null
        }
        zookeeper = new ZooKeeper(serverAddress, sessionTimeout,
            new Watcher {
                def process(event: WatchedEvent) {
                    sessionEvent(assignLatch, event)
                }
            })
        assignLatch.countDown()
        logger.info("Attempting to connect to cloud servers {}", serverAddress)
        connectionLatch.await()
    }

    //only for test session expired
    private[utils] def buildAnotherSession = {
        val self = new CountDownLatch(1)
        val zk = new ZooKeeper(serverAddress, sessionTimeout, new Watcher {
            def process(event: WatchedEvent) {
                event.getState match {
                    case KeeperState.SyncConnected =>
                        //monitor
                        logger.debug("Connected to cloud server")
                        self.countDown()
                    case other =>
                        logger.debug("other state:{}", event.getState)
                }

            }
        }, zookeeper.getSessionId, zookeeper.getSessionPasswd)
        self.await()

        zk
    }

    def sessionEvent(assignLatch: CountDownLatch, event: WatchedEvent) {
        logger.info("Receive event: {}", event)
        assignLatch.await()
        event.getState match {
            case KeeperState.SyncConnected =>
                logger.info("Connected to cloud server")
                runInNoThrow {
                    connectedFun(this)
                }
                //session过期引起
                if (sessionExpiredFlag.intValue() == 1) {
                    watchers.values().foreach(fn => runInNoThrow {
                        executeInZookeeper(fn)
                    })
                    sessionExpiredFlag.set(0)
                }
                connectionLatch.countDown()
            case KeeperState.Expired =>
                logger.info("session expired,so reconnect")
                sessionExpiredFlag.incrementAndGet()
                connect
            case KeeperState.Disconnected =>
                logger.info("lost connection to cloud server")

            case _ => // Disconnected -- zookeeper library will handle reconnects
        }
    }

    /**
     * 创建一个路径
     * @param path 创建的路径
     * @param createMode 创建的模式
     * @param data 创建路径的数据
     */
    def createPath(path: String, createMode: CreateMode = CreateMode.PERSISTENT, data: Option[String] = None) {
        createPathWithData(path, createMode, if (data.isDefined) Some(data.get.getBytes("UTF-8")) else None)
    }

    /**
     * 创建一个Ephemeral的path
     * @param path 路径
     * @param createMode 创建的模式
     * @param data 数据
     * @return 创建结果
     */
    def createEphemeralPath(path: String, createMode: CreateMode = CreateMode.EPHEMERAL, data: Option[String] = None) = {
        zookeeper.create(path,
            if (data.isDefined) data.get.getBytes("UTF-8") else null,
            Ids.OPEN_ACL_UNSAFE,
            createMode)
    }

    /**
     * 关闭zookeeper
     */
    def close {
        executeInZookeeper(_.close())
    }

    /**
     * zookeeper回调
     * @param fun zookeeper中的操作函数
     * @return 函数的返回结果
     */
    def executeInZookeeper[T](fun: ZooKeeper => T) = fun(zookeeper)

    /**
     * 监听某一路径的子路径
     * @param path 监视路径
     * @param fun 子路径发生变化时候的处理函数
     */
    def watchChildren(path: String, fun: (Seq[String]) => Unit,pathOnDeletedFun:Option[Function1[String,Unit]]=None): Unit = {
        //得到状态,必须是存在的节点
        val someStat = stat(path)
        if (someStat.isEmpty) {
            throw new IllegalStateException("path doesn't exit")
        }
        val lastStat = someStat.get

        //访问子节点数据的异步函数
        val children2Callback = new Children2Callback {
            def processResult(rc: Int, path: String, ctx: AnyRef, children: java.util.List[String], stat: Stat) {
                //通过contex中获得传入的stat和watcher
                val arr = ctx.asInstanceOf[Array[AnyRef]]
                val paramStat = arr(0).asInstanceOf[Stat]
                val paramWatcher = arr(1).asInstanceOf[Watcher]
                //如果stat不为空，则更新lastStat
                if (stat != null) DataTree.copyStat(stat, lastStat)
                //检查子节点版本，如果不一致，则再Watch，一直到版本号一样
                if (paramStat != null && stat != null && stat.getCversion != paramStat.getCversion) {
                    logger.debug("children version not equals,so rewatch")
                    selfWatchChildren(path, paramWatcher, this, stat)
                }
                else // 运行对应的业务逻辑
                    runInNoThrow({
                        fun(children)
                    })
            }
        }

        //子节点的Watcher
        val childrenWatcher = new Watcher {
            def process(event: WatchedEvent) = {
                event.getType match {
                    case EventType.NodeChildrenChanged =>
                        //捕获到节点数据发生变化
                        logger.debug("watched children....")
                        selfWatchChildren(path, this, children2Callback, lastStat)
                    case EventType.NodeDeleted =>
                        logger.debug("path {} deleted ",path)
                        //删除本节点的时候，执行对应的函数
                        if(pathOnDeletedFun.isDefined) pathOnDeletedFun.get.apply(path)
                    case other =>
                        logger.debug("other children event:{}", other)
                }
            }
        };
        //创建子节点Watch函数
        val watcherFunction = selfWatchChildren(path, childrenWatcher, children2Callback, lastStat)
        watchers.putIfAbsent("children_" + path, watcherFunction)
    }

    private def selfWatchChildren(path: String, nativeWatcher: Watcher, children2Callback: Children2Callback, lastStat: Stat): (ZooKeeper => Unit) = {
        val watcher = (zk: ZooKeeper) => {
            logger.debug("lastStat:{},watcher:{}", lastStat, nativeWatcher)
            zk.getChildren(path, nativeWatcher, children2Callback, Array(lastStat, nativeWatcher))
        }
        executeInZookeeper(watcher)

        watcher
    }

    /**
     * 查看节点的数据
     * @param path 节点路径
     * @param fun 节点数据发生变化时候的响应函数
     */
    def watchNodeData(path: String, fun: (Option[String]) => Unit): Unit = {
        //得到状态值
        val someStat = stat(path)
        if (someStat.isEmpty) {
            throw new IllegalStateException("path doesn't exit")
        }
        val lastStat = someStat.get
        val dataCallback = new DataCallback {
            def processResult(rc: Int, path: String, ctx: AnyRef, data: Array[Byte], stat: Stat) {
                //从上下文中获取对应的数据
                val arr = ctx.asInstanceOf[Array[AnyRef]]
                val paramStat = arr(0).asInstanceOf[Stat]
                val paramWatcher = arr(1).asInstanceOf[Watcher]
                if (stat == null) {
                    //如果没有stat，再次请求
                    selfWatchNode(path, paramWatcher, this, stat)
                    return
                }
                //更新lastStat
                DataTree.copyStat(stat, lastStat)
                //检查子节点版本，如果不一致，则再Watch，一直到版本号一样
                if (paramStat != null && stat != null && stat.getVersion != paramStat.getVersion) {
                    logger.debug("node version not equals,so rewatch")
                    selfWatchNode(path, paramWatcher, this, stat)
                }
                else {
                    if (data != null) {
                        runInNoThrow {
                            fun(Some(new String(data, "UTF-8")))
                        }
                    }
                }
            }
        }
        val watcher = new Watcher {
            def process(event: WatchedEvent) = event.getType match {
                case EventType.NodeDataChanged =>
                    selfWatchNode(path, this, dataCallback, lastStat)
                case EventType.NodeDeleted =>
                    runInNoThrow {
                        fun(None)
                    }
                    watchers.remove("node_" + path)
                case other =>
                    logger.debug("other node event:{}", other)
            }
        }

        val watcherFunction = selfWatchNode(path, watcher, dataCallback, lastStat)
        watchers.put("node_" + path, watcherFunction)
    }

    private def selfWatchNode(path: String, nativeWatcher: Watcher, dataCallback: DataCallback, lastStat: Stat): (ZooKeeper => Unit) = {
        val watcher = (zk: ZooKeeper) => {
            zk.getData(path, nativeWatcher, dataCallback, Array(lastStat, nativeWatcher))
        }
        executeInZookeeper(watcher)

        watcher
    }

    /**
     * 创建路径
     * @param path zookeeper中的路径
     * @param createMode 创建的模式
     * @param data 创建的数据
     */
    def createPathWithData(path: String, createMode: CreateMode, data: Option[Array[Byte]]) {
        for (path <- subPaths(path, '/')) {
            try {
                zookeeper.create(path, data.getOrElse(null), Ids.OPEN_ACL_UNSAFE, createMode)
            } catch {
                case nee: NodeExistsException =>
                //do nothing
            }
        }
    }

    /**
     * Given a string representing a path, return each subpath
     * Ex. subPaths("/a/b/c", "/") == ["/a", "/a/b", "/a/b/c"]
     */
    private def subPaths(path: String, sep: Char) = {
        val l = path.split(sep).toList
        val paths = l.tail.foldLeft[List[String]](Nil) {
            (xs, x) =>
                (xs.headOption.getOrElse("") + sep.toString + x) :: xs
        }
        paths.reverse
    }

    /**
     * 得到某一路径的状态
     * @param path 路径
     * @return 路径状态值
     */
    def stat(path: String): Option[Stat] = {
        try {
            val stat = zookeeper.exists(path, false)
            if (stat == null) None else Some(stat)
        } catch {
            case e: ConnectionLossException =>
                None
        }
    }

    /**
     * 获取节点路径的内容
     * @param path 节点路径
     * @param encoding 编码 默认为 UTF-8
     * @return 路径值
     */
    def get(path: String, encoding: String = "UTF-8"): Option[String] = {
        try {
            val data = zookeeper.getData(path, false, null)
            if (data == null) None else Some(new String(data, encoding))
        } catch {
            case e: NoNodeException =>
                None
        }
    }

    /**
     * 设置文本内容
     * @param path 节点路径
     * @param data 节点的数据
     * @param encoding 编码
     */
    def set(path: String, data: String, encoding: String = "UTF-8") {
        zookeeper.setData(path, data.getBytes(encoding), -1)
    }

    /**
     * 删除路径
     * @param path 节点路径
     */
    def delete(path: String) {
        try {
            zookeeper.delete(path, -1)
        } catch {
            case e: NoNodeException =>
            //ingore
        }
    }

    /**
     * 得到某一路径的子路径集合
     * @param path 节点路径
     * @return 子节点集合
     */
    def getChildren(path: String): Seq[String] = {
        try {
            zookeeper.getChildren(path, false)
        } catch {
            case e: NoNodeException =>
                Seq[String]()
        }
    }

    /**
     * 循环删除某一个目录，包含子目录
     * @param path 节点路径
     */
    def deleteRecursive(path: String) {
        val children = getChildren(path)
        for (node <- children) {
            deleteRecursive(path + '/' + node)
        }
        delete(path)
    }

    /**
     * 关闭客户端
     */
    def shutdown {
        try {
            zookeeper.close()
        } catch {
            case e =>
        }
    }
    def runInNoThrow(fun: =>Unit){
        try{fun}catch{case e=> logger.error(e.getMessage,e)}
    }
}